from ._internal.frameworks.spacy import load
from ._internal.frameworks.spacy import save
from ._internal.frameworks.spacy import projects
from ._internal.frameworks.spacy import load_runner
from ._internal.frameworks.spacy import load_project

__all__ = ["projects", "load", "load_project", "load_runner", "save"]
